/*
Change Log. 10 March 2009
	Add: the ELLBETA option to the PSF 'level' parameter, which would be useful for testing the 2-d psf work. 
	Version: 0.2
*/

#include "PsfState.h"
#include "CalServer.h"
#include <iostream>
#include <string>
#include <sstream>

void PsfState::setValueFromPara( int ev_p ) {
    stringstream ios;
    double temp;
    Instrument id = (Instrument) Xmm::toInstrument(stringParameter("instrument"));

    _instrument.set(id);
    switch (id) {
    case Xmm::EMOS1:
    case Xmm::EMOS2:
      _state = new CalStateEmos(NULL);
      break;
    case Xmm::EPN:
      _state = new CalStateEpn(NULL);
      break;
    case Xmm::RGS1:
    case Xmm::RGS2:
      _state = new CalStateRgs(NULL);
      break;
    case Xmm::OM:
      _state = new CalStateOm(NULL);
      break;
    case Xmm::ERM:
    case Xmm::SC:
      _state = new CalState(NULL);
      break;
    default:
      cerr << "Illegal Instrument: " << stringParameter("instrument") << endl;
      goto ERROR;
    }

#define PIFACTIVE(k) if ( _state->k->isActive() )
#define PSETVALUE(k, str, type) PIFACTIVE(k) _state->k->set(type##Parameter(#str))
#define PSETVALUEEX(k, str) PIFACTIVE(k) _state->k->set(str)

    PSETVALUE(ccd(), ccdnr, int);
    PSETVALUEEX(node(), Xmm::toNode(stringParameter("node")));
    PSETVALUEEX(mode(),Xmm::toMode(stringParameter("cammode")));
    PSETVALUEEX(filter(), Xmm::toFilter(stringParameter("filter")));
    PSETVALUE(temperature(), ccdtemp, real);
    PSETVALUE(temperatureCamera(), camtemp, real);
    PSETVALUE(temperatureTracking(), temptrack, boolean);
    PSETVALUEEX(exposureStartTime(), (Xmm::missionReferenceTime + realParameter("starttime")));
    PSETVALUE(exposureDuration(), duration, real);

    {
       AccuracyLevel level;
       string sLevel(stringParameter("level"));

       if (!sLevel.compare("LOW")) level = LOW;
       else if (!sLevel.compare("MEDIUM")) level = MEDIUM;
       else if (!sLevel.compare("EXTENDED")) level = EXTENDED;
       else if (!sLevel.compare("HIGH")) level = HIGH;
       else if (!sLevel.compare("ELLBETA")) level = ELLBETA;
       else {
         cerr << "Illegal Level: " << sLevel << endl;
         goto ERROR;
       }

       // checking the current version libcal support the level or not, mainly for checking ELLBETA
       const vector<AccuracyLevel> & accus = calServer.state()->accuracyLevel()->allowedValues();
       unsigned j=0;
       for(; j < accus.size() && level != accus[j]; j++) ;
       if (j >= accus.size()) {
         cerr << "The cal library does not support this accuracy level: " << toString(level) << ". Please try linking psfgen to higher version of cal library." << endl;
         goto ERROR;
       }

       PSETVALUEEX(accuracyLevel(),level);
    }

    PSETVALUE(auxiliaryParameter(), parameter, int);
    PSETVALUE(randomize(), random, boolean);
    PSETVALUE(ocb(), binning, int);


    _ev.set(realParameter("energy", ev_p));
    ios << realParameter("theta") << endl;
    ios >> temp;
    _theta.set(temp);
    ios << realParameter("phi") << endl;
    ios >> temp;
    _phi.set(temp / 180.0 * M_PI);
    ios << realParameter("rotate") << endl;
    ios >> temp;
    _rotate.set(temp / 180.0 * M_PI);
//    _xangle.set(realParameter("xangle") / 180.0 * M_PI);
//    _yangle.set(realParameter("yangle") / 180.0 * M_PI);
    _coor.set(CoorType::toType(stringParameter("coortype")));
    _xsize.set(intParameter("xsize"));
    _ysize.set(intParameter("ysize"));
    _file.set(stringParameter("output"));

#undef PSETVALUE
#undef PSETVALUEEX
#undef PIFACTIVE
    return;

ERROR:
    if (_state) delete _state;
    _state = NULL;
    //throw new xception("Illegal state values. Can't create state.");
}

ostream &operator << (ostream &os, const PsfState &state)
{
#define PRINTVAR(var) \
  if (state.var()->isActive()) os<<state.var()->get(); \
  else os << "*inactive*"
#define GET(var) (state.var()->isActive() ? state.var()->get() : -1000)


  os << "   Instrument       : "; PRINTVAR(instrument); os << endl;
  if (state.state())
      os << state.state()[0] << endl;
  else
      os << "  ( NULL state)" << endl;
  os << "   Energy (eV)      : "; PRINTVAR(ev); os << endl;
  os << "   Theta (arcsec)   : "; PRINTVAR(theta); os << endl;
  os << "   Phi (arc)        : "; PRINTVAR(phi); os << endl;
  os << "   Rotate (arc)     : "; PRINTVAR(rotate); os << endl;
printf("%0.8f -- %0.8f -- %0.8f\n", GET(theta), GET(phi)*180/M_PI, GET(rotate)*180/M_PI);
//  os << "   XAngle           : "; PRINTVAR(xangle); os << endl;
//  os << "   YAngle           : "; PRINTVAR(yangle); os << endl;
  os << "   Coor Type        : "; PRINTVAR(coor); os << endl;
  os << "   xsize (pixel)    : "; PRINTVAR(xsize); os << endl;
  os << "   ysize (pixel)    : "; PRINTVAR(ysize); os << endl;
  os << "   file             : "; PRINTVAR(file); os << endl;
#undef PTINTVAR

  return os;
}

