/*
 psfgen

   Authors:  Tao Song - University of Leicester
         
   Description: Generate PSF from ccf
 
   Comments: 
 
  $Id: psfgen.cc,v 1.1 2011/01/11 16:36:10 rsaxton Exp $
*/

#include "psfgen.h"

#include <Tracer.h>
#include <errstr.h>

#include "PsfState.h"
#include "CalServer.h"
#include "DataSubspace.h"
#include "param.h"
#include "PsfRegion.h"
#include "ASCregion.h"
#include "PsfPosition.h"
#include "STime.h"

typedef double	real;
typedef bool	boolean;
typedef vector<double>	realList;

#define GETVALUE(name, type)    type name = type##Parameter(#name);
#define SETVALUE(name, value)          setParameter(#name, value);
#define SHOWSTR(name)	{GETVALUE(name, string); cout << #name << " = " << name << endl;}
#define SHOWINT(name)	{GETVALUE(name, int); cout << #name << " = " << name << endl;}
#define SHOWBOOL(name)	{GETVALUE(name, boolean); cout << #name << " = " << (name ? "yes" : "no") << endl;}
#define SHOWREAL(name)	{GETVALUE(name, real); cout << #name << " = " << name << endl;}
#define SHOWTIME(name)	{GETVALUE(name, real); cout << #name << " = " << STime(Xmm::missionReferenceTime + name) << endl;}
#define SHOWLIST(name)	{GETVALUE(name, realList); cout << #name << " = [ "; for (realList::iterator i = name.begin(); i < name.end(); ++i) cout << *i << ", "; cout << "]\n";}

void ShowParameters()
{
    
	SHOWSTR(instrument);
	SHOWSTR(image);
	SHOWBOOL(useodf);
	SHOWSTR(region);
	SHOWINT(ccdnr);
	SHOWSTR(node);
	SHOWSTR(cammode);
	SHOWSTR(filter);
	SHOWREAL(ccdtemp);
	SHOWREAL(camtemp);
	SHOWBOOL(temptrack);
	SHOWTIME(starttime);
	SHOWREAL(duration);
	SHOWSTR(level);
	SHOWINT(parameter);
	SHOWBOOL(random);
	SHOWINT(binning);
	SHOWSTR(ccf);
	SHOWBOOL(ccflog);
	SHOWLIST(energy);
	SHOWLIST(weight);
	SHOWREAL(theta);
	SHOWREAL(phi);
	SHOWREAL(rotate);
	SHOWINT(xsize);
	SHOWINT(ysize);
	SHOWSTR(output);
}

void psfgen ()
//int main()
{
PROC
   // get the image parameter to see whether an image is supplied
   string fitsfile = stringParameter("image");

   unsigned count = parameterCount("energy");
   char buff[50]; buff[49] = 0;
   
   PsfPosition psfPos;
   CoverFBox box;

   // if image parameter is given, read parameters from image dataset
   if (!fitsfile.empty()) { // image dataset supplied
      psfPos.setDataSet(fitsfile, stringParameter("region"));

      //overwrite instrument if not supplied in parameter
      GETVALUE(instrument, string);
      if (instrument.empty()) SETVALUE(instrument, psfPos.instrument());

      //overwrite starttime and duration if not supplied in parameter
      GETVALUE(duration, real);
      if (duration == 0) {
         STime start(psfPos.obsStart()), end(psfPos.obsEnd());
         SETVALUE(starttime, (start - Xmm::missionReferenceTime));
         SETVALUE(duration, (end - start));
      }

      //The following parameters can be detected by region, but because psfPos is initialised with region expression already, 
      //the parameters of psfPos should correspond to the region if there is a region string. So these parameters can be 
      //overwritten by the values of psfPos.
      box = psfPos.coverBox();
//cout << box; 
//cout << psfPos;
      
      GETVALUE(ccdnr, int);
      if (ccdnr == 0) SETVALUE(ccdnr, psfPos.ccdnr());

//	cout << box.refPoint() << endl;
      GETVALUE(theta, real);
      if (theta < 0.0) SETVALUE(theta, box.refPoint().x());
      GETVALUE(phi, real);
      if (phi < 0.0) SETVALUE(phi, box.refPoint().y());
      GETVALUE(rotate, real);
      if (rotate == -1.0 ) SETVALUE(rotate, box.pa());
      GETVALUE(coortype, string);
      if (coortype.empty()) SETVALUE(coortype, CoorType::toString(box.coor()));
//      GETVALUE(xangle, real);		// version 1.1.4 or older
//      if (xangle < 0.0) SETVALUE(xangle, box.xAngle());		// version 1.1.4 or older
//      GETVALUE(yangle, real);		// version 1.1.4 or older
//      if (yangle < 0.0) SETVALUE(yangle, box.yAngle());		// version 1.1.4 or older
      
	// The actual pixel_right - pixel_left = box_width, pixel_bottom - pixel_top = box_height.
	// Therefore the width in pixel and the height in pixel should both be increased by 1 in order to match the image size
      GETVALUE(xsize, int);
      if (xsize == 0) //SETVALUE(xsize, static_cast<int>(box.xWidth() / 22.0 + 0.5));
         SETVALUE(xsize, static_cast<int>(box.xOrgDetSize() / 22.0 + 1.5));
      GETVALUE(ysize, int);
      if (ysize == 0) //SETVALUE(ysize, static_cast<int>(box.yWidth() / 22.0 + 0.5));
         SETVALUE(ysize, static_cast<int>(box.yOrgDetSize() /22.0 + 1.5));
   } else { // NO image dataset supplied
      string regstr = stringParameter("region");
      if (!regstr.empty()) {
		psfPos.setDataSet(NULL, stringParameter("region"));
	      	box = psfPos.coverBox();
      }

      //if no way to detect instrument, default it to "EMOS1"
      GETVALUE(instrument, string);
      if (instrument.empty()) SETVALUE(instrument, "EMOS1");
      
      //if no way to detect duration, default it to 3000.0
      GETVALUE(duration, real);
      if (duration == 0) SETVALUE(duration, 3000.0);

      //the following parameters can be detected by region
      //GETVALUE(region, string);
         
      //if no way to detect ccd no, default it to 1
      GETVALUE(ccdnr, int);
      if (ccdnr == 0) SETVALUE(ccdnr, (regstr.empty() ? 1 : psfPos.ccdnr()));

      //if no way to detect theta, phi, default it to 0
      GETVALUE(theta, real);
      if (theta < 0.0) SETVALUE(theta, (regstr.empty() ? 0.0 : box.refPoint().x()));
      GETVALUE(phi, real);
      if (phi < 0.0) SETVALUE(phi, (regstr.empty() ? 0.0 : box.refPoint().y()));
      GETVALUE(rotate, real);
      if (rotate == -1.0 && box.yWidth() > 0) SETVALUE(rotate, box.pa());
      GETVALUE(coortype, string);
      if (coortype.empty()) SETVALUE(coortype, (regstr.empty() ? "XY" : CoorType::toString(box.coor())));
//      GETVALUE(xangle, real);		// version 1.1.4 or older
//      if (xangle < 0.0) SETVALUE(xangle, 0.0);	//version 1.1.4 or older
//      GETVALUE(yangle, real);		// version 1.1.4 or older
//      if (yangle < 0.0) SETVALUE(yangle, 0.0);	// version 1.1.4 or older

      //if no way to detect xsize, ysize, default it to 199
      GETVALUE(xsize, int);
      if (xsize == 0) SETVALUE(xsize, (regstr.empty() ? 199 : static_cast<int>(box.xOrgDetSize() / 22.0 + 1.5)));
      GETVALUE(ysize, int);
      if (ysize == 0) SETVALUE(ysize, (regstr.empty() ? 199 : static_cast<int>(box.yOrgDetSize() / 22.0 + 1.5)));
   }


   // Get all the default parameters
   PsfState state;
   state.setValueFromPara();
   string psffile = state.file()->get();

   // if no output, return the theta, phi, rotation, xsize, ysize
   if (psffile.empty()) {
//      cout << "instrument=" << stringParameter("instrument") << " theta=" << state.theta()->get() << " phi=" << realParameter("phi") << " rotate=" << realParameter("rotate") << " xangle=" << realParameter("xangle") << " yangle=" << realParameter("yangle") << " xsize=" << state.xsize()->get() << " ysize=" << state.ysize()->get() << endl;	// version 1.1.4 or older
      cout << "instrument=" << stringParameter("instrument") << " theta=" << state.theta()->get() << " phi=" << realParameter("phi") << " rotate=" << realParameter("rotate") << " coortype=" << stringParameter("coortype") << " xsize=" << state.xsize()->get() << " ysize=" << state.ysize()->get() << endl;

      return;
   }

   // Need to generate a PSF but without given energy
   if (!count) {
     errstr << "psfgen.cc: energy value missing!!"
	<< Sas::error("InvalidEneryValue");
     return;
   }

   if (state.state() == NULL) {
     errstr << "psfgen.cc: Cannot create a state!!"
	<< Sas::error("InvalidInstState");
     return;
   }

   DataSet *dataSet = dataSetServer->open(psffile, Dal::Create);

//   const XmmInstrument * oldInstrument = calServer.instrument();
   
   calServer.setInstrument(state.instrument()->get());
   calServer.pushState();
   calServer.atomServer()->restart();
   calServer.setState(state.state()[0]);

   string ccf = stringParameter("ccf");
   if (ccf.length()) {
	errstr << "use user own ccf : "
		<< ccf
		<< Sas::warning("UserOwnCCF");
	calServer.setCcf(ccf);
   }

   PsfDataServer *pds = 0; pds = calServer.getAtom(pds);
   std::vector<unsigned long> extents(2);
   extents[0] = state.ysize()->get();  //rows of expected PSF image
   extents[1] = state.xsize()->get();  //cols of expected PSF image

   Array *Normal = NULL;
   ArrayData *Normald = NULL;
   real32 *dNormal = NULL;

   for (unsigned i = 0; ; ) {
      Psf &psf = pds->psf(state.ev()->get(), state.theta()->get(), state.phi()->get());
      
      if (realParameter("rotate") >= 0) psf.setPosAngle(state.rotate()->get()); 
      else errstr << "rotate is not used"
		<< Sas::warning("InvalidPA");

      // set the correct size for image windows
      
      float win_col, win_row;

      switch (state.state()->accuracyLevel()->get())
      {
        case LOW:
        case MEDIUM:
	case HIGH:
			win_row = (extents[0] + 1) / 50.0, win_col = (extents[1] + 1) /50.0;
//			if (box.yWidth() == 0) win_row = (extents[0] + 1) / 50.0, win_col = (extents[1] + 1) /50.0;
//			else win_row = static_cast<int>(box.yWidth() / 22.0 + 1) / 50.0, win_col = static_cast<int>(box.xWidth() / 22.0 + 1) / 50.0;
			break;
	// psf.image() makes different size PSF images for different level.
        case EXTENDED: 
	case ELLBETA:
 			win_row = extents[0] / 55.0, win_col = extents[1] / 55.0;
// 			if (box.yWidth() == 0) win_row = extents[0] / 55.0, win_col = extents[1] / 55.0;
//			else win_row = static_cast<int>(box.yWidth() / 22.0) / 55.0, win_col = static_cast<int>(box.xWidth() / 22.0) / 55.0;
			break;
        default: 	errstr << "psfgen.cc: Unrecognised accuracy level!!"
				<< Sas::error("InvalidAccLevel");
			return;
      }
      
/********************************************************************************************************************************
*																*
*	Be Noticed:														*
*	The output of psf.image() may be still problematic. Ideally, if the output were directly saved in a fits file and shown	*
*	in ds9, it will match the output of calview and follows the this rule (at least): increasing phi causes it look like	*
*	rotating clockwise. Unfortunately, this is not. The reason for this problem is due to inconsitstance between the format	*
*	for representing an image, i.e. column interlace for psf.image() output and line interlace for fits. In order to get a	*
*	correct output, a TRANSPOSE must be done when copying contents between these two objects. That is copy the Nrd line from*
*	Image<X> to the Nrd column of fits. Accordingly, when xsize ne ysize, the output dimension must swap.			*
*																*
********************************************************************************************************************************/
      double win = max(win_col, win_row); 
      Image<float> image = psf.image(SpatialWindow<double>(win, win));

      std::vector<unsigned long> realext(2);
      realext[0] = image.yExtent();   //rows of real PSF image
      realext[1] = image.xExtent();   //cols of real PSF image

assert(realext[0] >= extents[0]);
assert(realext[1] >= extents[1]);
      // when energy is a list more than one value, create the primary image as the combination of all subsets
      
      if (Normal == NULL && count > 1) {
	Normal = dataSet->addArray("PRIMARY", Dal::Real32, extents);
	*(Normal->addAttribute("ENERGY", "energy value of the PSF")) = "Summation";
	Normald = Normal->data();
	dNormal = Normald->real32Data();
	unsigned nElements = Normal->elements();
	for (unsigned n=0; n<nElements; ++n) dNormal[n] = 0;
      }

      // when energy is a list has only one value, the primary image is the specified energy psf itself
      if (i==0 && count==1)
	sprintf(buff, "PRIMARY%c", 0);
      else
	sprintf(buff, "PSF_%f%c", realParameter("energy", i), 0);

      Array *a = dataSet->addArray(buff, Dal::Real32, extents);
      sprintf(buff, "%f%c", realParameter("energy", i), 0);
      *(a->addAttribute("ENERGY", "energy value of the PSF")) = buff;
      unsigned nElements = a->elements();

      ArrayData *ad = a->data();
      real32 *d = ad->real32Data();


      int nn=0;
      int startrow = (realext[0] - extents[0]) / 2;
      int startcol = (realext[1] - extents[1]) / 2;
//      if (state.xangle()->get() > 0) {	//version 1.1.4 or older
      if (state.coor()->get() == CoorType::XY) {
      	 for (unsigned iy=0; iy<extents[0]; ++iy) {
		 for (unsigned ix=0; ix<extents[1]; ++ix) {
             		d[nn++] = image[startcol+ix][startrow+iy];
         	}
	 }
      } else {
	 startcol--;
      	 for (unsigned iy=0; iy<extents[0]; ++iy) {
		 for (unsigned ix=extents[1]; ix>0; --ix) {
             		d[nn++] = image[startcol+ix][startrow+iy];
         	}
	 }
      }


      if (dNormal) {
	unsigned n = parameterCount("weight");
	float w = realParameter("weight", i<n ? i : n-1);
	for (unsigned n=0; n<nElements; ++n) dNormal[n] += d[n]*w;
      }
    
      if (++i >= count) break;
      
      state.setValueFromPara(i);
   } 
   
   if (Normald) delete Normald;

   dataSetServer->close(dataSet);

   if (booleanParameter("ccflog")) errstr << calServer.ccfAccessLog()
					<< Sas::warning("CCFLog");

   calServer.popState();

ENDPROC
}
